// Vertikal-Generator
// fr Philips BRE-Modul 
// Frank Brennecke, 12.01.2013
// Version 1.1
// Fr ATMega88 mit interner 8MHz-Steuerung
// Fuses: Extended F9, High: DF, Low: 62

// horizontale Ausgangsfrequenzen an C0: 37,9 kHz (VGA800), 31,4 kHz (VGA640), 15,625 kHz (FBAS)
// Zustzliche Ausgangsfrequenzen bei internem Modus: fx1, fx2, fx4, fx8 bei Taktverhltnis 1:1

// vertikale Ausgangsfrequenzen an C1: 60 Hz (VGA800), 60 Hz (VGA640), 50 Hz (FBAS)
// Der Vertikalgenerator bentigt aufgrund der geringen internen Frequenzen keinen eigenen Quarz-Oszillator, 
//   sondern kann mit der internen Frequenzeinheit betrieben werden. Beim Betrieb als interner Generator 
//   wird er automatisch mit dem Horizontalgenerator synchronisiert. Er besitzt daher keinen eigenen internen Modus.

#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/io.h>

//vertical Outputs

#define vsync_on PORTC = 0x01; // C0 positiver Sync
#define vsync_off PORTC	= 0x00;

// Modi fr externe Sync-Signale
#define VGA800 		0x1C	// SVGA 800x600 selected - passend fr PORTB/1
#define VGA640 		0x1A 	// VGA 640x480 selected - passend fr PORTB/2
#define FBAS 		0x16 	// FBAS selected - passend fr PORTB/3
// Interner Generator - diese Betriebsart macht aus der Schaltung entweder ein Video-Modul (analog zum EE2023) oder ein VGA-Modul 
#define INTGEN_VGA	0x0A	// Internal VGA Signal Generator active - passend fr PORTB/4
#define INTGEN_FBAS	0x06	// Internal FBAS Signal Generator active - passend fr PORTB/4

#define VGA800_IN	0x01	// SVGA 800x600 erkannt, Eingang D0
#define VGA640_IN	0x02 	// VGA 640x480 erkannt, Eingang D1
// FBAS-Signal wird erkannt, wenn D1 "0" ist
// wenn D0 und D1 "1" sind, wird SVGA statt VGA gewhlt
#define REVERSE_SYNC	0x20	// Reverse Sync, Eingang D5 
#define INTERN_GEN		0x40	// Interner Generator aktiviert, Eingang D6
#define DELAY_SWITCH	0x80	// Delay-Switch, Bildposition einstellen, Eingang D7
#define SYNC_INTERN		0x08	// Eingang fr internen Sync, D3

#define SYNC_VGA		0x40	// Eingang fr VGA-Sync, B6
#define SYNC_FBAS		0x80	// Eingang fr FBAS-Sync, B7

#define VGA640_lines_internal  	525 // VGA-Signal hat 525 Zeilen
#define FBAS_lines_internal	 	313 // FBAS-Signal hat 2x312,5 Zeilen

#define INT_CONTROL			EICRA
#define INT0_FALLING_EDGE	0x02 	//Bit 1 gesetzt 
#define INT0_RISING_EDGE	0x03	//Bit 0 und Bit 1 gesetzt
#define INT1_FALLING_EDGE	0x08	//Bit 3 gesetzt
#define INT1_RISING_EDGE	0x0C	//Bit 2 und Bit 3 gesetzt

#define INT_ENABLE			EIMSK
#define INT0_ENABLE			0x01	//Bit 0 gesetzt
#define INT1_ENABLE			0x02	//Bit 1 gesetzt
#define INT_DISABLE			0x00	//kein Bit gesetzt


//Global variables
volatile unsigned int linecount;// current line
volatile unsigned int sync_mode;   // Welcher Sync-Modus ist eingestellt?
volatile unsigned int reverse_sync; // reverse Sync ntig?
volatile unsigned int max_delay_time; // Maximale Verschiebung der Sync-Position (Bildlage)
volatile int porch_delay;  // Tatschliche Verschiebung der Sync-Position (Bildlage)
   
   
ISR(INT0_vect)
{
// Dieser Interrupt wird fr jedes Halbbild des FBAS-Signals einmal ausgelst (50 Hz)
	int i;
	
	cli();
	
	// Wert fr Bildlage bercksichtigen
	
	
//	for( i = porch_delay; i < max_delay_time; i++)
//	{
//		_delay_ms(5);
//	}
	
// vertical Sync	
	vsync_on;
    _delay_ms(25);
	vsync_off;	
	sei();
}   
   

ISR(INT1_vect)
{
// Dieser Interrupt wird fr jedes Halbbild des VGA-Signals einmal ausgelst (60 bzw. 56 Hz)

	int i;
	cli();	// Interrupts abschalten


	if (sync_mode == VGA800)
	{
// Wert fr Bildlage bercksichtigen	
//		for( i = porch_delay; i < max_delay_time; i++)
//		{
//			_delay_ms(20);
//		}
		vsync_on;
		_delay_ms(5);
		vsync_off;	
	}
	else
	{
// Wert fr Bildlage bercksichtigen	
//	    for( i = porch_delay; i < max_delay_time; i++)
//		{
//			_delay_ms(4);
//		}
		vsync_on;
		_delay_ms(15);
		vsync_off;
	}	
	

	sei();  // Interrupts wieder aktivieren

}





static void avr_init(void)
{
	
//Set power mode
  set_sleep_mode(SLEEP_MODE_IDLE); 	
  
//init Ports

// Port C: Output
// Der C-Port enthlt alle notwendigen Ausgangssignale
// C0 - Sync-Signal (Sync positiv)
  DDRC = 0xFF;
  PORTC = 0x00;
  
// Port D: Input
// Der D-Port enthlt alle notwendigen Eingnge
// D0 - Detect-Leitung VGA800
// D1 - Detect-Leitung VGA640
// D2 - Zeilen-Sync, INT0
// D3 - internes V-Sync-Signal
// D4 - frei
// D5 - Reverse Sync - ntig, wenn das Eingangssignal nicht der Norm entspricht
// D6 - interner Generator erzeugt VGA statt FBAS
// D7 - Bildlage-Taster

  DDRD = 0x00;
  PORTD = 0xFF;   // Pull-up-Widerstnde aktivieren	
  
// Port B: Output bei B1, B2, B3, B4
// Die B-Ports werden fr die LED-Anzeige des gewhlten Betriebsmodus genutzt (VGA800, VGA640, FBAS, intern)
// Sie dienen auerdem zur Zuschaltung von Ladekondensatoren der nachgeschalteten Sgezahnstufen, um bei
// verschiedenen Frequenzen annhernd gleiche Amplituden zu erreichen.
// B1 - VGA800
// B2 - VGA640
// B3 - FBAS
// B4 - interner Generator - immer zusammen mit B2 oder B3 aktiv
// B6 - VGA VSync-Signal
// B7 - FBAS VSync-Signal
  DDRB = 0x1E;  
//  PORTB = 0x00;
  
  return;
}



static void Setup_Sync(void)
{
	unsigned int input_values;
	unsigned int detected_sync_mode;
	unsigned int detected_reverse_sync;

// Daten von Port D lesen und auswerten
	input_values = PIND;
	
// Einstellung des Sync-Modes: Schalter D1 unterscheidet zwischen FBAS und VGA
// ist zustzlich Schalter D0 geschlossen, wird auf SVGA optimiert
	if (input_values & VGA640_IN) 
	{
		if (input_values & VGA800_IN) 
			detected_sync_mode = VGA800;
		else
			detected_sync_mode = VGA640;
	}
	else 
		detected_sync_mode = FBAS;
		
	
// ist der interne Grafikmodus aktiviert? 
// der interne Grafikmodus versetzt das Modul in die Betriebsart "Video-Modul EE3023" bzw. "VGA-Modul"		
	if (input_values & INTERN_GEN)
	{
		if (detected_sync_mode == FBAS)
		{
			detected_sync_mode = INTGEN_FBAS;		// Ausgnge fr VGA640 und intern sind aktiviert
		}
		else
		{
			detected_sync_mode = INTGEN_VGA;
		}	
	}
			
// besitzt das Sync-Signal umgekehrte Polaritt gegenber der Spezifikation 
	detected_reverse_sync = (input_values & REVERSE_SYNC);
	
// Hat eine Vernderung gegenber dem letzten Setup stattgefunden ?	
	if ((detected_sync_mode != sync_mode) | (detected_reverse_sync != reverse_sync))
	{
		cli();   // Interrupts abschalten, Signal bricht zusammen
		INT_ENABLE = 0;
		sync_mode = detected_sync_mode;
		reverse_sync = detected_reverse_sync;
			
		PORTB = sync_mode;
		
		switch (sync_mode)
		{
			case VGA800: 
			{
				porch_delay = 20;
				max_delay_time = 25; 
				if (reverse_sync)
					INT_CONTROL = INT1_RISING_EDGE; // INT0 rising edge, positive pulse 
				else
					INT_CONTROL = INT1_FALLING_EDGE; // INT0 rising edge, negative pulse
				INT_ENABLE = INT1_ENABLE;
				break;
			}
			
			case VGA640: 
			{
				porch_delay = 0;	
				max_delay_time = 25;
				if (reverse_sync)
					INT_CONTROL = INT1_FALLING_EDGE; // INT0 rising edge, positive pulse 
				else
					INT_CONTROL = INT1_RISING_EDGE; // INT0 rising edge, negative pulse
				INT_ENABLE = INT1_ENABLE;	
				break;
			}
		
			case FBAS: 
			{
				porch_delay = 15;
				max_delay_time = 25;
				INT_CONTROL = INT0_RISING_EDGE;
				INT_ENABLE = INT0_ENABLE;
				break;
			}
			
			default:
				break;
			
			
		}
	

		
	}

// enable Interrupts
	sei();	
	
// Bildverschiebung durch optionalen Teiler 
// Bildverschiebung wird hier geprft: wenn die taste gedrckt ist (Masse), wird der Wert um 1 erhht.	

	if (!(PORTB & DELAY_SWITCH))
	{
		porch_delay = porch_delay + 10;
		if (++porch_delay > max_delay_time)
		{
			porch_delay = -max_delay_time;   
		} 
	}
}		





int main(void)
{
	linecount = 0;
	sync_mode = 0x00;
	reverse_sync = 0x00;
	porch_delay = 0;
	max_delay_time = 25;
	
// Ports initialisieren
	avr_init();
	
	_delay_ms(1000);
	
//	PORTB = 0x00;
	
    while(1)
	{
		Setup_Sync();
    }
	return(0);
}





	
    